<?php
/**
 * @file
 * Helper functions for processing PHP.
 *
 * The goal is to translate from one language (PHP) to another (Twig). Semantic
 * analysis on an abstract syntax tree (AST) would be easiest for such
 * transformations, right?  However, there are no well-developed tools for
 * producing an AST from PHP in PHP.
 *
 * The PHP internals use C. Their "parser" also does a bit of semantic analysis
 * and code generation (see Parsekit). It doesn't quite work for our purposes.
 *
 * One of the main goals of Twig is to enable human-readable, easy-to-maintain
 * templates. Part and parcel of that goal is preserving whitespace and
 * comments. However, a traditional compiler/translator will ditch whitespace
 * and comments the first chance it gets; its goal it to produce machine code
 * (and not human-readable code).
 *
 * If we want to preserve some of the layout (whitespace and comments) of the
 * original PHP, we need an AST from a parser that preserves (or is aware of)
 * such things in the token stream. This requirement (along with bulding a PHP
 * parser in PHP) means we have to grow our tool ourselves. NB: a concrete
 * syntax tree would probably do just as well as an AST here.
 *
 * What follows is a crude lexer+. It weakly parses some elements of PHP, and
 * applies some semantic understanding to transform a few expressions from PHP
 * to Twig.  YMMV.
 */

// This was standalone code.  Renaming every function seems unwieldy
// so we namespace instead.
namespace TwigifyLuthor;

/**
 * Generates an error message in Twig comment format for unexpected tokens.
 *
 * @param array $token
 *   Token from parsed PHP.
 *
 * @return string
 *   A Twig format comment describing the unexpected token and its line number.
 */
function unexpected_token_message($token) {
  $token_name = token_name($token[0]);
  $token_text = strtr($token[1], array("\t" => '\t', "\n" => '\n'));
  $line_number = $token[2];
  $error_message = "ERROR: Encountered unexpected token $token_name on line $line_number: $token_text";
  return "{# $error_message #}";
}

/**
 * token_get_all() returns a list of tokens, but some tokens that are single
 * characters get returned raw (and not as an array of name, token, lineNumber).
 * We do three things:
 *  - Convert all singlets to arrays;
 *  - ensure a closing tag if we're in the middle of a PHP block at EOF (valid
 *   PHP doesn't require one);
 *  - Change the list to a stack.
 *
 * @see token_get_all()
 */
define("T_OPEN_PARENTHESIS", 4913);
define("T_CLOSE_PARENTHESIS", 4914);
define("T_COMMA", 4915);
define("T_SEMICOLON", 4916);
define("T_OPEN_BRACKET", 4917);
define("T_CLOSE_BRACKET", 4918);
define("T_QUESTION", 4919);
define("T_COLON", 4920);
define("T_BOOLEAN_NOT", 4921);
define("T_LESS_THAN", 4922);
define("T_GREATER_THAN", 4923);
define("T_ASSIGNMENT", 4924);
define("T_PLUS", 4925);
define("T_MINUS", 4926);
define("T_CONCAT", 4927);
define("T_MUL", 4928);
define("T_DIV", 4929);
define("T_MOD", 4930);
define("T_BIT_NOT", 4931);
define("T_BIT_AND", 4932);
define("T_BIT_XOR", 4933);
define("T_BIT_OR", 4934);
define("T_IGNORE_ERROR", 4935);
define("T_OPEN_BRACE", 4936);
define("T_CLOSE_BRACE", 4937);
define("T_REFERENCE", 4932);
define("T_AMPERSAND", 4932);

/**
 * Normalises output of token_get_all() so each element is an array.
 *
 * @param array $token_list
 *   An array of token identifiers.
 *
 * @return array
 *   An array of token identifiers.
 *
 * @see token_get_all()
 */
function preprocess_tokens($token_list) {
  $token_types = array(
    '(' => T_OPEN_PARENTHESIS,
    ')' => T_CLOSE_PARENTHESIS,
    ',' => T_COMMA,
    ';' => T_SEMICOLON,
    '[' => T_OPEN_BRACKET,
    ']' => T_CLOSE_BRACKET,
    '?' => T_QUESTION,
    ':' => T_COLON,
    '!' => T_BOOLEAN_NOT,
    '<' => T_LESS_THAN,
    '>' => T_GREATER_THAN,
    '=' => T_ASSIGNMENT,
    '+' => T_PLUS,
    '-' => T_MINUS,
    '.' => T_CONCAT,
    '*' => T_MUL,
    '/' => T_DIV,
    '%' => T_MOD,
    '~' => T_BIT_NOT,
    '&' => T_BIT_AND,
    // Sigh, depends on context: binary "&" is T_BIT_AND where unary "&" is
    // T_REFERENCE.
    '^' => T_BIT_XOR,
    '|' => T_BIT_OR,
    '@' => T_IGNORE_ERROR,
    '{' => T_OPEN_BRACE,
    '}' => T_CLOSE_BRACE,
  );
  $result = array();
  $line_number = 1;
  foreach ($token_list as $token) {
    if (gettype($token) !== 'array') {
      $token = array(
        $token_types[$token],
        $token,
        // Get from previous token.
        // @todo could be off if previous token has trailing whitespace that
        // contains line breaks.
        $line_number,
      );
    }
    else {
      $line_number = $token[2];
    }
    array_push($result, $token);
  }
  if ($token[0] !== T_CLOSE_TAG && $token[0] !== T_INLINE_HTML) {
    array_push($result, array(
      T_CLOSE_TAG,
      '?>',
      // Get from previous token (which is guaranteed to exist if valid PHP).
      $line_number,
    ));
  }
  $result = array_reverse($result);
  return $result;
}

/**
 * Filters whitespace from a list of tokens.
 *
 * This is done in place (destructively); note the pass by reference.
 *
 * @param array $token_list
 *   A token list returned by token_get_all() or preprocess_tokens().
 *
 * @return array
 *   An array of token definitions, minus the whitespace token.
 *
 * @see token_get_all()
 * @see preprocess_tokens()
 */
function filter_whitespace(&$token_list) {
  $token_list = array_filter(
    $token_list,
    function ($item) {
      // Kill whitespace but keep all others.
      return ($item[0] === T_WHITESPACE) ? FALSE : TRUE;
    }
  );
}

/**
 * Grabs a block of PHP source code off the top of the stack, and returns it.
 *
 * The first block of php between opening (<?php) and closing (?>) tags is
 * returned.
 *
 * Note the pass by reference; the incoming stack is modified as the PHP block
 * is popped.
 *
 * @param array $token_stack
 *   An array of tokenized PHP.
 *
 * @return array
 *   An array of PHP tokens.
 *
 * @see token_get_all()
 * @see preprocess_tokens()
 */
function pop_php(&$token_stack) {
  $result = array();
  // First token is always T_OPEN_TAG, so pop right away.
  do {
    $token = array_pop($token_stack);
    array_push($result, $token);
  } while ($token[0] !== T_CLOSE_TAG);
  return array_reverse($result);
}

/**
 * Converts an array of tokens into PHP source code, wrapped in a Twig comment.
 *
 * @param array $token_stack_to_print
 *   An array of tokenized PHP.
 *
 * @return string
 *   The PHP source wrapped in Twig comment tags instead of <?php tags.
 */
function print_php($token_stack_to_print) {
  $result = '';
  while ($token = array_pop($token_stack_to_print)) {
    $type = $token[0];
    $name = $token[1];
    switch ($type) {
      case T_OPEN_TAG:
        // PHP open tag to Twig open comment (and preserve whitespace).
        $name = strtr($name, array('<?php' => '{# PHP:'));
        break;

      case T_CLOSE_TAG:
        // PHP close tag to Twig close comment (and preserve whitespace).
        $name = strtr($name, array('?>' => '#}'));
        break;

      default:
        // Pass through, but make sure not to close the Twig comment by mistake.
        $name = strtr($name, array('#}' => '#hash}brace'));
        break;
    }
    $result .= $name;
  }
  return $result;
}

/**
 * Breaks a block of PHP up into pieces.
 *
 * The pieces aren't always statements. Opening if/elseif/else, foreach and
 * comments are peeled off. We return an array of token stacks, one entry per
 * subblock.
 *
 * @param array $token_stack
 *   An array of tokenized PHP.
 *
 * @return array
 *   An array of tokenized PHP, grouped by subblock.
 */
function php_subblocks($token_stack) {
  // A place to store the returned subblocks.
  $subblocks = array();

  // The first token is always T_OPEN_TAG.
  $open_token = array_pop($token_stack);
  // Preserve any trailing whitespace for the first subblock.
  $whitespaces = explode('<?php', $open_token[1]);
  $first_whitespace = $whitespaces[1];
  $is_first = TRUE;
  $line_number = $open_token[2];

  do {

    // The first token is always T_OPEN_TAG.
    $subblock = array(
      array(
        T_OPEN_TAG,
        '<?php' . ($is_first ? $first_whitespace : ''),
        $line_number,
      ),
    );
    $is_first = FALSE;

    // Find the first non-whitespace token.
    do {
      $token = array_pop($token_stack);
      $type = $token[0];
      array_push($subblock, $token);
    } while ($type === T_WHITESPACE);

    // Comments, close braces, and endif/endforeach go in their own subblock as
    // does the opening for if/elseif/else and foreach.
    // LIE: everybody else is delimited by a semicolon or a T_CLOSE_TAG
    // @todo the above is wishful thinking; there are a gajillion cases to
    // handle here, and really what we need is the parse tree (which'd identify
    // blocks and statements); heck, even "for" and "switch" statements break
    // this, as do  closures -- all of which aren't unexpected, even in
    // templates.
    switch ($type) {
      case T_COMMENT:
      case T_DOC_COMMENT:
        // Do nothing; the comment has already been pushed, and whitespace is
        // picked up below.
        break;

      case T_ENDIF:
      case T_ENDFOREACH:
        // Look ahead for the optional semicolon.
        // If present, gobble it; if not, put back the look-ahead token.
        $look_ahead_token = array_pop($token_stack);
        $look_ahead_type = $look_ahead_token[0];
        if ($look_ahead_type === T_SEMICOLON) {
          array_push($subblock, $look_ahead_token);
        }
        else {
          array_push($token_stack, $look_ahead_token);
        }
        break;

      case T_CLOSE_BRACE:
        // Look ahead.
        // If we see the optional semicolon, that's the end of the block: stop.
        // If we see anything but an else/elseif, we just finished an if or
        // foreach: stop and put back lookahead.
        // If we see an elseif, fall through to below to find the conditional.
        // If we see an else, we'd like to jump to the T_ELSE case below (but
        // goto doesn't allow that so we repeat code instead).
        // @todo ugly but necessary; is there a cleaner way to do this?
        $whitespace_stack = array();
        do {
          $token = array_pop($token_stack);
          $type = $token[0];
          array_push($whitespace_stack, $token);
        } while ($type === T_WHITESPACE);
        $look_ahead_token = array_pop($whitespace_stack);
        $look_ahead_type = $look_ahead_token[0];
        if ($look_ahead_type === T_SEMICOLON) {
          foreach ($whitespace_stack as $token) {
            array_push($token_stack, $token);
          }
          array_push($subblock, $look_ahead_token);
          break;

        }
        elseif ($look_ahead_type === T_ELSE) {
          // Whitespace gets ignored in an else subblock.
          array_push($subblock, $look_ahead_token);
          // Look for the opening brace or colon.
          do {
            $token = array_pop($token_stack);
            $type = $token[0];
            array_push($subblock, $token);
          } while ($type !== T_OPEN_BRACE);
          break;

        }
        // FIXME??? does "else if" get converted to an T_ELSEIF token?
        elseif ($look_ahead_type === T_ELSEIF) {
          // Whitespace gets ignored in an elseif subblock.
          array_push($subblock, $look_ahead_token);
          // No break.
        }
        else {
          array_push($token_stack, $look_ahead_token);
          while ($whitespace_stack) {
            array_push($token_stack, array_pop($whitespace_stack));
          }
          break;

        }
      case T_IF:
      case T_ELSEIF:
      case T_FOREACH:
        // Look for the opening parenthesis on the condition / loop array.
        do {
          $token = array_pop($token_stack);
          $type = $token[0];
          array_push($subblock, $token);
        } while ($type !== T_OPEN_PARENTHESIS);
        // Look for the matching close parenthesis.
        $open_paren_count = 1;
        do {
          $token = array_pop($token_stack);
          array_push($subblock, $token);
          if ($token[0] === T_OPEN_PARENTHESIS) {
            $open_paren_count++;
          }
          elseif ($token[0] === T_CLOSE_PARENTHESIS) {
            $open_paren_count--;
          }
        } while ($open_paren_count > 0);
        // No break; fall through to T_ELSE to get the brace or colon.

      case T_ELSE:
        // Look for the opening brace or colon.
        do {
          $token = array_pop($token_stack);
          $type = $token[0];
          array_push($subblock, $token);
        } while ($type !== T_OPEN_BRACE && $type !== T_COLON);
        break;

      default:
        // Look for a semicolon or the close tag.
        // NB: $type is from above; we check first because we may have hit an
        // empty block.
        while ($type !== T_SEMICOLON && $type !== T_CLOSE_TAG) {
          $token = array_pop($token_stack);
          $type = $token[0];
          array_push($subblock, $token);
        }
        // If we stopped on the close tag, put it and any preceding whitespace
        // back on.
        if ($type === T_CLOSE_TAG) {
          $close_token = array_pop($subblock);
          array_push($token_stack, $close_token);
          $look_back_token = array_pop($subblock);
          if ($look_back_token[0] === T_WHITESPACE) {
            array_push($token_stack, $look_back_token);
          }
          else {
            array_push($subblock, $look_back_token);
          }
        }
        break;
    }

    // Soak up whitespace to postpend to the T_CLOSE_TAG.
    $whitespace_stack = array();
    do {
      $token = array_pop($token_stack);
      $type = $token[0];
      array_push($whitespace_stack, $token);
    } while ($type === T_WHITESPACE);

    // We went one token too far; push it back and add T_CLOSE_TAG (unless it
    // was a T_CLOSE_TAG; then we just stop).
    if ($type !== T_CLOSE_TAG) {
      // Push back the last token; it's the start of the next block.
      $next_block_token = array_pop($whitespace_stack);
      array_push($token_stack, $next_block_token);
      // Add a T_CLOSE_TAG to the subblock, and glue the whitespace onto it.
      $whitespace = '';
      foreach ($whitespace_stack as $token) {
        $whitespace .= $token[1];
      }
      // Update for T_OPEN_TAG at top of loop.
      $line_number = $token[2];
      array_push($subblock, array(
        T_CLOSE_TAG,
        '?>' . $whitespace,
        $line_number,
      ));
    }
    else {
      // Push whitespace_stack onto the subblock.
      foreach ($whitespace_stack as $token) {
        array_push($subblock, $token);
      }
      // NB: token_stack is exhausted; we don't need to update line_number.
    }

    // Add the subblock to the return.
    $subblocks[] = array_reverse($subblock);

  } while ($token_stack);

  // Return what we found.
  return $subblocks;
}


/**
 * Functions to parse and transform variable globs. A glob could be a:
 *  - variable name
 *  - object access modifier (public, private, protected)
 *  - object method
 *  - object member
 *  - array
 *  - array key
 */

/**
 * Grabs a glob off the top of the stack, and returns a Twigified version.
 *
 * Note the pass by reference; the incoming stack is modified as the variable
 * glob is popped.
 *
 * @param array $token_stack
 *   An array of tokenized PHP.
 *
 * @return string
 *   The name of the glob (for example className.methodName).
 *
 * @see _gobble_variable_recurse()
 */
function gobble_variable(&$token_stack) {
  // Tokens that we might see in a variable glob.
  // NB: basically we stop on T_COMMA, an unbalanced T_CLOSE_PARENTHESIS or
  // T_CLOSE_BRACKET, T_SEMICOLON, or T_CLOSE_TAG there are other possibilities,
  // but those are the ones we understand for now.
      
  $accept_token_types = array(
    // Variable names.
    T_VARIABLE,
    // Object access.
    T_OBJECT_OPERATOR,
    // Object methods.
    T_OPEN_PARENTHESIS,
    T_CLOSE_PARENTHESIS,
    // Object methods and members.
    T_STRING,
    // Array access.
    T_OPEN_BRACKET,
    T_CLOSE_BRACKET,
    // Array keys.
    T_CONSTANT_ENCAPSED_STRING,
    T_LNUMBER,
    // Added T_REFERENCE (4932) to address https://www.drupal.org/node/2444753
    T_REFERENCE, 
    // Array keys.
    // @todo T_ENCAPSED_AND_WHITESPACE, T_NUM_STRING, T_STRING_VARNAME

  );
  // A place to hold the tokens we're gobbling.
  $var_tokens = array();

  // Gobble, gobble; NB: first token is expected to be T_VARIABLE.
  $bracket_count = 0;
  $parenthesis_count = 0;
  do {
    $token = array_pop($token_stack);

    array_push($var_tokens, $token);
    // Make sure not to leave the tree node we started from.
    if ($token[0] === T_OPEN_PARENTHESIS) {
      $parenthesis_count++;
    }
    if ($token[0] === T_CLOSE_PARENTHESIS) {
      $parenthesis_count--;
    }
    if ($token[0] === T_OPEN_BRACKET) {
      $bracket_count++;
    }
    if ($token[0] === T_CLOSE_BRACKET) {
      $bracket_count--;
    }
  } while (in_array($token[0], $accept_token_types, TRUE) &&
    $bracket_count >= 0 &&
    $parenthesis_count >= 0);
  // We went one token too far; put it back.
  // (which is OK because there's always T_CLOSE_TAG, T_CLOSE_PARENTHESIS or
  // T_CLOSE_BRACKET, or whatever other non-accept token we hit).
  array_push($token_stack, array_pop($var_tokens));
  // @todo Unnecessary if we assume valid PHP, but we could check that the
  // counts are both 0, or one 0 and one -1 (but no other state).
  // The T_VARIABLE that started is at bottom of the stack; reverse the order.
  $var_tokens = array_reverse($var_tokens);

  // Start a recursive parse.
  return _gobble_variable_recurse('', $var_tokens);
}

/**
 * Recursively walk the token stack, processing globs as we go.
 *
 * Will parse PHP variables, object member, method accesses, and array accesses.
 *
 * Object access:
 *   - A variable as member/method name needs attribute(object,method/member)
 *   - A method with arguments requires attribute(object,method,arguments)
 *   - Methods with no arguments and T_STRING-identified members are easy.
 * Array access:
 *   - array accesses with constant strings are easy
 *   - variable array access needs attribute(array,item)
 * @todo We're gonna ignore variables that store functions for now; we will,
 *      however, correctly catch variables that store object methods.
 *
 * @param string $parent_name
 *   The name of the opening token.
 * @param array $var_tokens
 *   A list of tokens in the glob being parsed.
 *
 * @return string
 *   The ancestor opening token for this glob.
 */
function _gobble_variable_recurse($parent_name, $var_tokens) {
  $next_token = array_pop($var_tokens);
  if (!$next_token) {
    return $parent_name;
  }
  $type = $next_token[0];
  $name = $next_token[1];

//  if($type==){};
  // @FIXME: Extremely ugly hardcoding of parser token to its concrete value instead of corresponding PHP constant. 
  if ($type=='4932'){
    $next_token = array_pop($var_tokens);
    $type = $next_token[0];
    $name = $next_token[1];
  }

  switch ($type) {
    case T_REFERENCE:
      // The variable is being passed by reference. 
      // We need to ignore this token, since we pass all variables by reference in our new functions. 
      // This is also not being used, since we bypass it by checking for T_REFERENCE right before our switch-case.
      return;
    case T_VARIABLE:
      // We must be at the start; ignore the parent.
      // @FIXME: We must be at the start?" Isn't this where the pass-by-reference bug is happening? 
      return _gobble_variable_recurse(substr($name, 1), $var_tokens);

    case T_OBJECT_OPERATOR:
      // Object access.
      // Member or method of object.
      $mom_token = array_pop($var_tokens);
      // T_VARIABLE or T_STRING (either can be a member or method name).
      $mom_type = $mom_token[0];
      $mom_name = $mom_token[1];
      if ($mom_type === T_VARIABLE) {
        $mom_name = substr($mom_name, 1);
      }
      $look_ahead_token = array_pop($var_tokens);
      if (!$look_ahead_token) {
        // We hit the end; return the appropriate accessor.
        switch ($mom_type) {
          case T_STRING:
            return "$parent_name.$mom_name";

          case T_VARIABLE:
            return "attribute($parent_name,$mom_name)";

          default:
            return unexpected_token_message($mom_token);
        }
      }
      // Keep going an analyze lookAhead.
      $look_ahead_tpe = $look_ahead_token[0];
      switch ($look_ahead_tpe) {
        case T_OPEN_PARENTHESIS:
          $look_ahead_two_token = array_pop($var_tokens);
          if ($look_ahead_two_token[0] === T_CLOSE_PARENTHESIS) {
            // We hit an argumentless method.
            switch ($mom_type) {
              case T_STRING:
                $var_name = "$parent_name.$mom_name";
                break;

              case T_VARIABLE:
                $var_name = "attribute($parent_name,$mom_name)";
                break;

              default:
                return unexpected_token_message($mom_token);
            }
            return _gobble_variable_recurse($var_name, $var_tokens);
          }
          else {
            // We hit a method with arguments; back up and recurse.
            array_push($var_tokens, $look_ahead_two_token);
            array_push($var_tokens, $look_ahead_token);
            return _gobble_variable_recurse("$parent_name,$mom_name", $var_tokens);
          }
          break;

        case T_OPEN_BRACKET:
          // We hit an array access; back up and recurse.
          array_push($var_tokens, $look_ahead_token);
          switch ($mom_type) {
            case T_STRING:
              $var_name = "$parent_name.$mom_name";
              break;

            case T_VARIABLE:
              $var_name = "attribute($parent_name,$mom_name)";
              break;

            default:
              return unexpected_token_message($mom_token);

          }
          return _gobble_variable_recurse($var_name, $var_tokens);

        default:
          return unexpected_token_message($look_ahead_token);
      }
      break;

    case T_OPEN_BRACKET:
      // Array access.
      $look_ahead_token = array_pop($var_tokens);
      $look_ahead_tpe = $look_ahead_token[0];
      $look_ahead_name = $look_ahead_token[1];
      switch ($look_ahead_tpe) {
        case T_LNUMBER:
          // Pop the next token and check it's a closing bracket.
          $look_ahead_two_token = array_pop($var_tokens);
          if ($look_ahead_two_token[0] !== T_CLOSE_BRACKET) {
            return unexpected_token_message($look_ahead_two_token);
          }
          // Process the lookAhead token and recurse on the rest.
          return _gobble_variable_recurse($parent_name . "[$look_ahead_name]", $var_tokens);

        case T_CONSTANT_ENCAPSED_STRING:
          // Pop the next token and check it's a closing bracket.
          $look_ahead_two_token = array_pop($var_tokens);
          if ($look_ahead_two_token[0] !== T_CLOSE_BRACKET) {
            return unexpected_token_message($look_ahead_two_token);
          }
          // Process the lookAhead token and recurse on the rest.
          $look_ahead_name = trim($look_ahead_name, '\'"');
          return _gobble_variable_recurse("$parent_name.$look_ahead_name", $var_tokens);

        case T_VARIABLE:
          // Grab the variable glob.
          array_push($var_tokens, $look_ahead_token);
          $var_name = gobble_variable($var_tokens);
          // Pop the next token and check it's a closing bracket.
          $look_ahead_two_token = array_pop($var_tokens);
          if ($look_ahead_two_token[0] !== T_CLOSE_BRACKET) {
            return unexpected_token_message($look_ahead_two_token);
          }
          // Process the lookAhead token and recurse on the rest.
          return _gobble_variable_recurse("attribute($parent_name,$var_name)", $var_tokens);

        default:
          return unexpected_token_message($look_ahead_token);
      }
      break;

    case T_OPEN_PARENTHESIS:
      // We hit a method with arguments; gobble up the arguments.
      $args = array();
      $stop = FALSE;
      do {
        $args[] = gobble_variable($var_tokens);
        $next_token = array_pop($var_tokens);
        // Arguments should be separated by commas and delimited by parentheses.
        if ($next_token[0] !== T_COMMA && $next_token[0] !== T_CLOSE_PARENTHESIS) {
          return unexpected_token_message($next_token);
        }
        // Either comma or close parenthesis.
        $stop = ($next_token[0] === T_CLOSE_PARENTHESIS);
      } while (!$stop);
      return _gobble_variable_recurse("attribute($parent_name,$var_name," . implode(',', $args) . ')', $var_tokens);

    default:
      return unexpected_token_message($next_token);

  }
  return FALSE;
}

/**
 * Processing expressions.
 */

/**
 * Processes a boolean expression into an equivalent Twig expression.
 *
 * Both variables and logical operators are parsed. The function expects the
 * expression to be wrapped in parentheses (as for "if" statements).
 *
 * @param array $token_stack
 *   An array of tokenized PHP.
 *
 * @return string
 *   The Twig equivalent of the PHP logical expression.
 */
function gobble_boolean_expression(&$token_stack) {

  // A place to hold the translation of the tokens we're gobbling.
  $result = '';

  // Gobble, gobble; NB: first token is expected to be T_OPEN_PARENTHESIS.
  $parenthesis_count = 0;
  do {
    $token = array_pop($token_stack);
    $type = $token[0];
    switch ($type) {
      case T_OPEN_PARENTHESIS:
        $parenthesis_count++;
        // If we're in the middle of the expression, add another parenthesis.
        if ($parenthesis_count > 1) {
          $result .= '(';
        }
        break;

      case T_CLOSE_PARENTHESIS:
        $parenthesis_count--;
        // Remove any trailing whitespace (may or may not be present).
        $result = trim($result);
        // If we're in the middle of the expression, add another parenthesis.
        if ($parenthesis_count > 0) {
          $result .= ')';
        }
        break;

      case T_VARIABLE:
        array_push($token_stack, $token);
        $result .= gobble_variable($token_stack) . ' ';
        break;

      case T_LESS_THAN:
        $result .= '< ';
        break;

      case T_GREATER_THAN:
        $result .= '> ';
        break;

      case T_IS_SMALLER_OR_EQUAL:
        $result .= '<= ';
        break;

      case T_IS_GREATER_OR_EQUAL:
        $result .= '>= ';
        break;

      case T_IS_EQUAL:
        $result .= '== ';
        break;

      case T_IS_NOT_EQUAL:
        $result .= '!= ';
        break;

      // @todo case T_IS_IDENTICAL: case T_IS_NOT_IDENTICAL
      // @todo Unsupported in Twig?
      case T_BOOLEAN_AND:
      case T_LOGICAL_AND:
        $result .= 'and ';
        break;

      case T_BOOLEAN_NOT:
        $result .= 'not ';
        break;

      case T_BOOLEAN_OR:
      case T_LOGICAL_OR:
        $result .= 'or ';
        break;

      // @todo case T_LOGICAL_XOR
      // @todo Twig only has b-xor); implement as (x or y) and not (x and y)?
      case T_DNUMBER:
        $result .= $token[1] . ' ';
        break;

      case T_LNUMBER:
        $result .= $token[1] . ' ';
        break;

      case T_CONSTANT_ENCAPSED_STRING:
        $result .= $token[1] . ' ';
        break;

      // Function and function-like calls.
      case T_STRING:
      case T_EMPTY:
      case T_ISSET:
        $result .= $token[1];
        break;

      default:
        $result .= unexpected_token_message($token);
        break;
      // @todo Should we accept arithmetic (+ - * / % ~ & ^ | << >>) and casts?
      // @todo assignments?  -- and ++?  what else?
    }
  } while ($token_stack && $parenthesis_count > 0);

  // There should always be at least a T_CLOSE_TAG left.
  // ALT: if token_stack is empty *and* we didn't end on a close paren, then
  // throw an error.
  if (count($token_stack) == 0) {
    return "{# ERROR: Unexpected end of token stream at line {$token[2]} #}";
  }

  return $result;
}


/**
 * Returns the parameters from a call to theme().
 *
 * We assume that a T_STRING has already been popped off the stack with text
 * "theme".
 *
 * @param array $token_stack
 *   the token stack after T_STRING has been popped off; passed by value
 *
 * @return array
 *   An array containing one member with the following keys:
 *    - hook: the first argument to the theme() call, the hook name.
 *    - variables: (optional) an array of variables passed to theme().
 */
function twig_theme_call_parameters($token_stack) {
  // First token should be an open parenthesis for the function call.
  $paren_token = array_pop($token_stack);
  if ($paren_token[0] !== T_OPEN_PARENTHESIS) {
    return array('hook' => 'ERROR: missing open parenthesis to theme() call');
  }

  // Look for the first argument, the hook name.
  // Though this could be any expression, for simplicity we'll only deal with
  // constant strings for now.
  do {
    $token = array_pop($token_stack);
  } while ($token[0] === T_WHITESPACE);
  if ($token[0] !== T_CONSTANT_ENCAPSED_STRING) {
    return array('hook' => 'ERROR: could not find hook name');
  }
  $hook_name = $token[1];

  // Look ahead: either we should find a comma (and a second argument) or a
  // close parenthesis.
  do {
    $token = array_pop($token_stack);
  } while ($token[0] === T_WHITESPACE);
  if ($token[0] === T_CLOSE_PARENTHESIS) {
    return array('hook' => $hook_name);
  }
  elseif ($token[0] !== T_COMMA) {
    return array('hook' => 'ERROR: missing a comma or a close parenthesis');
  }

  // The next argument should be the $variables array.
  $variables_array = _twig_array_tokens_to_array($token_stack);

  // Check for the closing parenthesis.
  do {
    $token = array_pop($token_stack);
  } while ($token[0] === T_WHITESPACE);
  if ($token[0] !== T_CLOSE_PARENTHESIS) {
    return array('hook' => 'ERROR: could not find close parenthesis');
  }

  // Success: return what we've found.
  return array('hook' => $hook_name, 'variables' => $variables_array);
}


/**
 * Looks for an array in the token stream and build up a PHP array from it.
 *
 * @param array $token_stack
 *   An array of tokenized PHP.
 *
 * @return array
 *   A PHP array as defined in the token stack.
 */
function _twig_array_tokens_to_array(&$token_stack) {
  // Look for T_ARRAY to start.
  do {
    $token = array_pop($token_stack);
  } while ($token[0] === T_WHITESPACE);
  if ($token[0] !== T_ARRAY) {
    return array('ERROR: could not find T_ARRAY');
  }
  // Followed by an open parenthesis.
  do {
    $token = array_pop($token_stack);
  } while ($token[0] === T_WHITESPACE);
  if ($token[0] !== T_OPEN_PARENTHESIS) {
    return array('ERROR: could not find array\'s open parenthesis');
  }

  // Run through looking for key=>value pairs.
  $result = array();
  do {
    // Constant string for the key.
    do {
      $token = array_pop($token_stack);
    } while ($token[0] === T_WHITESPACE);
    if ($token[0] !== T_CONSTANT_ENCAPSED_STRING) {
      return array('ERROR: could not find array key');
    }
    $key = $token[1];

    // => , or )
    do {
      $token = array_pop($token_stack);
    } while ($token[0] === T_WHITESPACE);

    switch ($token[0]) {
      case T_COMMA:
        // Key was value and array keeps going.
        $result[] = $key;
        break;

      case T_CLOSE_PARENTHESIS:
        // Key was value and array stops.
        $result[] = $key;
        // Switch and while.
        break 2;

      case T_DOUBLE_ARROW:
        // Do nothing and continue below looking for the value.
        break;

      default:
        return array('ERROR: unexpected token in array');
    }

    // Look for the value.
    do {
      $token = array_pop($token_stack);
    } while ($token[0] === T_WHITESPACE);
    if ($token[0] === T_ARRAY) {
      // If the value is an array, then get the value recursively.
      array_push($token_stack, $token);
      $value = _twig_array_tokens_to_array($token_stack);
      $result[$key] = $value;
      // Look for a comma or close parenthesis.
      do {
        $token = array_pop($token_stack);
      } while ($token[0] === T_WHITESPACE);
      if ($token[0] !== T_COMMA && $token[0] !== T_CLOSE_PARENTHESIS) {
        return array('ERROR: array ended abnormally');
      }
    }
    else {
      // Gobble up everything up to a comma or close parenthesis (while watching
      // for balanced parentheses).
      $value = $token[1];
      $open_parentheses = 0;
      do {
        $token = array_pop($token_stack);
        if ($token[0] === T_COMMA) {
          break;
        }
        elseif ($token[0] === T_OPEN_PARENTHESIS) {
          $value .= $token[1];
          $open_parentheses++;
        }
        elseif ($token[0] === T_CLOSE_PARENTHESIS) {
          if ($open_parentheses > 0) {
            $value .= $token[1];
            $open_parentheses--;
          }
          else {
            break;
          }
        }
        else {
          $value .= $token[1];
        }
      } while (TRUE);
      $result[$key] = $value;
    }

    // Keep going until the array closes.
  } while ($token[0] !== T_CLOSE_PARENTHESIS);

  // Return what we found.
  return $result;
}


/**
 * Rules for transforming PHP to Twig.
 */

/**
 * Checks to see if a PHP string matches a rule, returning a string if so.
 *
 * Rules are defined in rules.inc. Rules define the tokens that describe a
 * particular glob. Examples include 'empty', 'elseif', and 'print($var)'.
 *
 * @param array $rule
 *   A rule as defined in rules.inc.
 * @param array $statement_token_stack
 *   A tokenised PHP statement.
 *
 * @return mixed
 *   A Twig format glob of PHP code if the rule is matched; FALSE if not.
 */
function rule_match($rule, $statement_token_stack) {
  filter_whitespace($statement_token_stack);
  $vars = array();
  foreach ($rule['pattern'] as $type => $content) {
    $token = array_pop($statement_token_stack);
    if (!$token) {
      return FALSE;
    }
    if (!(
      ($token[0] === $type) ||
      ($token[0] === T_OPEN_PARENTHESIS && $type === T_BOOLEAN_EXPR)
    )
    ) {
      return FALSE;
    }
    if ($type === T_STRING && $content !== $token[1]) {
      return FALSE;
    }
    // Take special actions for some token types.
    switch ($type) {
      case T_VARIABLE:
        array_push($statement_token_stack, $token);
        $vars[] = gobble_variable($statement_token_stack);
        break;

      case T_BOOLEAN_EXPR:
        array_push($statement_token_stack, $token);
        $vars[] = gobble_boolean_expression($statement_token_stack);
        break;

      case T_COMMENT:
      case T_DOC_COMMENT:
        // Make sure we don't close the Twig comment.
        $vars[] = strtr($token[1], array('#}' => '#hash}brace'));
        break;

      case T_CONSTANT_ENCAPSED_STRING:
        $vars[] = $token[1];
        break;

      case T_OPEN_TAG:
        // Only used for comments; we need it at the end of the $vars array to
        // avoid dinging every other rule.
        $whitespaces = explode('<?php', $token[1]);
        $lead_whitespace = $whitespaces[1];
    }
  }
  if (count($statement_token_stack) > 0) {
    return FALSE;
  }
  else {
    if (isset($lead_whitespace)) {
      $vars[] = $lead_whitespace;
    }
    $name = vsprintf($rule['format'], $vars);
    // Merde! close tags can sometimes slurp up trailing whitespace.
    if ($token[0] === T_CLOSE_TAG) {
      $whitespaces = explode('?>', $token[1]);
      $name .= $whitespaces[1];
    }
    return $name;
  }
}

/**
 * Evaluate all our pattern matching rules on a block.
 *
 * Rules are defined in rules.inc. Rules define the tokens that describe a
 * particular glob. Examples include 'empty', 'elseif', and 'print($var)'.
 *
 * @param array $statement_token_stack
 *   Array of tokens from a PHP statement.
 *
 * @return string
 *   The substituted text of the PHP statement.
 *
 * @see rules.inc
 */
function apply_rules($statement_token_stack) {
  // Get $self_contained_rules, $open_brace_rules, and $close_brace_rules.
  require 'rules.inc';

  // Try self_contained_rules first.
  foreach ($self_contained_rules as $rule) {
    $print_rule = rule_match($rule, $statement_token_stack);
    // We found a match.
    if ($print_rule !== FALSE) {
      break;
    }
  }
  // We found a match.
  if ($print_rule !== FALSE) {
    return $print_rule;
  }

  // Track open braces.
  static $open_brace_stack = array();

  // Try open_brace_rules.
  foreach ($open_brace_rules as $rule) {
    $print_rule = rule_match($rule, $statement_token_stack);
    // We found a match.
    if ($print_rule !== FALSE) {
      break;
    }
  }
  if ($print_rule !== FALSE) {
    // We found a match; ifs and for[each]s increase the brace count.
    switch (substr($print_rule, 0, 5)) {
      case '{% if':
        array_push($open_brace_stack, T_IF);
        break;

      case '{% for':
        array_push($open_brace_stack, T_FOR);
        break;

      // default: elseif/else; do nothing.
    }
    return $print_rule;
  }

  // Try close_brace_rules.
  foreach ($close_brace_rules as $rule) {
    $print_rule = rule_match($rule, $statement_token_stack);
    // We found a match.
    if ($print_rule !== FALSE) {
      break;
    }
  }
  if ($print_rule !== FALSE) {
    // We found a match.
    $token_type = array_pop($open_brace_stack);
    return strtr($print_rule,
      ($token_type === T_IF) ?
        array('end' => 'endif') :
        array('end' => 'endfor')
    );
  }

  // No match on any rule.
  return FALSE;
}


/**
 * "Public" methods that can be called from anywhere else safely.
 */

/**
 * Takes a string of PHP and returns a string of Twig.
 *
 * @param string $php_string
 *   A string of PHP code.
 *
 * @return string
 *   A string of Twig syntax.
 */
function php_string_to_twig_string($php_string) {
  // Get a stream of tokens from the PHP string.
  $tokens = token_get_all($php_string);

  // Keep track of the output string.
  $result = '';

  // Do our thang.
  $token_stack = preprocess_tokens($tokens);
  while ($token = array_pop($token_stack)) {
    if ($token[0] === T_OPEN_TAG) {
      array_push($token_stack, $token);
      $php_token_stack = pop_php($token_stack);
      foreach (php_subblocks($php_token_stack) as $php_block_stack) {
        $statement_match = apply_rules($php_block_stack);
        $result .= ($statement_match !== FALSE) ? $statement_match : print_php($php_block_stack);
      }
    }
    elseif ($token[0] === T_INLINE_HTML) {
      $result .= $token[1];
    }
    else {
      $result .= unexpected_token_message($token);
      continue;
    }
  }

  // Return the twigified PHP.
  return $result;
}


/**
 * Converts a theme hook function body to a preprocess function.
 *
 * NB: We assume that the result of using reflection to get the function body
 * via file() gives us this function's body, and it alone.  That is, we match
 * both:
 *   $body =~ /^function $oldFunctionName ($variablesName) {/
 *   $body =~ /}\s*$/  <=  where the closing brace is the closing brace for this
 *   function (and not, say the closing brace for a one-line function that
 *   shares the same line as the closing brace for the function of interest;
 *   it's unusual fo shizzle, but is valid PHP).
 *
 * Actually, the second requirement isn't quite necessary; we stop processing on
 * a balanced close brace.  Also note we needn't mind return $foo vs
 * return($foo); the substitution of return with an assignment works correctly
 * regardless of enclosing parentheses.
 *
 * For example, to convert bartik_field__taxonomy_term_reference, call this
 * function with $old_theme_name = 'bartik' and $hook_name =
 * 'field__taxonomy_term_reference'.
 *
 * @todo Keep an array of const_strings we encounter, and a second array of
 * variablesName references; output a block of suggested moves to template
 * elementary suggestions (fixed) such as ifs, foreaches, and print/echos
 * rules for assignment, and concat assignment?
 * function to process string globs (including interpolation)?
 *
 * @param string $body
 *   Text of old theme hook function. Includes declaration and closing brace.
 * @param string $hook_name
 *   The Drupal hook we're supposed to be tinkering with.
 * @param string $old_theme_name
 *   The name of the (old) source theme for this function.
 * @param string $new_theme_name
 *   The name of the new (Twig) theme we're creating.
 *
 * @return array
 *   An array with one member with the following keys:
 *   'function_text' => text of the function declaration (with comment
 *                      doc blocks) to insert in template.php
 *   'function_name' => name of the new template preprocess hook
 *   'template_name' => file name for the template to hold the result
 *   'template_token' => array key in $variables used to hold output from the
 *                       preprocessor
 */
function twig_build_preprocess_function($body, $hook_name, $old_theme_name, $new_theme_name) {
  $old_function_name = $old_theme_name . '_' . $hook_name;
  // template_preprocess is not needed here.
  // $new_function_name = $new_theme_name . '_template_preprocess_' . $hook_name;
  // @TODO: Also, neet to add logic for checking to see if label_hidden is set in variables.
  $new_function_name = $new_theme_name . '_' . $hook_name;
  $new_template_name = strtr($hook_name, array('_' => '-')) . '.html.twig';
  // Or set as you please.
  $template_token = $new_function_name;

  // script tag needed for token_get_all
  $body = '<?php ' . $body;
  $body_tokens = token_get_all($body);

  // @FIXME: (No 2nd argument.)
  $token_stack = preprocess_tokens($body_tokens, FALSE);
  $output = '';

  // Strip the script tag we just used to tokenize.
  $script_tag = array_pop($token_stack);
  if ($script_tag[1] !== '<?php ') {
    $output .= '/* ' . 'EXPECTING SCRIPT TAG instead of ' . unexpected_token_message($function_token) . ' */';
  }

  // Grab the function keyword.
  $function_token = array_pop($token_stack);

  if ($function_token[0] !== T_FUNCTION) {
    $output .= '/* ' . unexpected_token_message($function_token) . ' */';
  }

  // Grab up to the old function's name.
  do {
    $token = array_pop($token_stack);
  } while ($token[0] === T_WHITESPACE);
  if ($token[0] !== T_STRING || $token[1] !== $old_function_name) {
    $output .= '/* ' . unexpected_token_message($token) . ' */';
  }

  // Grab the old function's arguments; store the name it uses for $variables.
  do {
    $token = array_pop($token_stack);
  } while ($token[0] === T_WHITESPACE);
  if ($token[0] !== T_OPEN_PARENTHESIS) {
    $output .= '/* ' . 'EXPECTING T_OPEN_PARENTHESIS instead of ' . unexpected_token_message($token) . ' */';
  }
  $arguments = array();
  $stop = FALSE;
  // @todo This will get tripped up by (an inappropriate) pass by reference to
  // the old function.
  do {
    $arguments[] = gobble_variable($token_stack);
    $next_token = array_pop($token_stack);

    // Arguments should be separated by commas and delimited by parentheses.
    if ($next_token[0] !== T_COMMA && $next_token[0] !== T_CLOSE_PARENTHESIS) {
      $output .= '/* ' . unexpected_token_message($next_token) . ' */';
    }
    $stop = ($next_token[0] === T_CLOSE_PARENTHESIS) || empty($token_stack);
  } while (!$stop);
  if (count($token_stack) == 0 || count($arguments) != 1) {
    $output .= '/* ' . unexpected_token_message($next_token) . ' */';
  }
  $variables_name = $arguments[0];

  $output = 'function ' . $new_function_name . '(&$' . $variables_name . ')';

  // Grab the function body.
  // First grab up to the opening brace.
  do {
    $token = array_pop($token_stack);
    $output .= $token[1];
  } while ($token[0] === T_WHITESPACE);
  if ($token[0] !== T_OPEN_BRACE) {
    $output .= '/* ' . unexpected_token_message($token) . ' */';
  }
  // Next grab up to the balancing closing brace.
  $open_brace_count = 1;
  $drupal_render_count = 0;
  $t_count = 0;
  $constant_string_count = 0;
  $variables_ref_count = 0;
  $theme_calls = array();
  do {
    $token = array_pop($token_stack);
    // Accumulate the output.
    switch ($token[0]) {
      case T_OPEN_BRACE:
        $open_brace_count++;
        $output .= $token[1];
        break;
      case T_CLOSE_BRACE:
        $open_brace_count--;
        $output .= $token[1];
        break;
      case T_RETURN:
        $output .= '$' . $variables_name . "['" . $template_token . "'] =";
        break;
      default:
        $output .= $token[1];
        break;
    }
    // Check for things to move to the associated template.
    switch ($token[0]) {
      case T_STRING:
        switch ($token[1]) {
          case 'drupal_render':
            $drupal_render_count++;
            break;

          case 't':
            $t_count++;
            break;

          case 'theme':
            $theme_calls[] = twig_theme_call_parameters($token_stack);
            break;
        }
        break;

      case T_CONSTANT_ENCAPSED_STRING:
        $constant_string_count++;
        break;

      case T_VARIABLE:
        if (substr($token[1], 1) === $variables_name) {
          $variables_ref_count++;
        }
        break;
    }
  } while ($open_brace_count > 0 && !empty($token_stack));
  if ($open_brace_count > 0 && empty($token_stack)) {
    $output .= '/* ' . unexpected_token_message($token) . ' */';
  }
  // We ignore everything after the closing brace (including the close tag).
  // Add a trailing newline (because we lost all whitespace after the closing
  // brace).
  $output .= "\n";

  // Add a docblock header documenting the new preprocess function.
  $header_doc_block = <<<EOHEADER
/**
 * Prepares variables for [description of $hook_name] templates.
 *
 * Default template: $new_template_name.
 *
 * @param array \$$variables_name
 *   An associative array containing:
 *   - $template_token: the markup outputted by the original $old_function_name theme function.
 */

EOHEADER;
  // NB: the last newline is ignored by the heredoc PHP syntax.
  // Add a docblock footer with suggestions for things to move to the template.
  $footer_doc_block_start = <<<EOFOOTERSTART
/**
 * The following could probably be moved to $new_template_name:

EOFOOTERSTART;
  // NB: The last newline is ignored by the heredoc PHP syntax.
  $footer_doc_block_middle = '';
  if ($drupal_render_count > 0) {
    $footer_doc_block_middle .= " *   $drupal_render_count call(s) to drupal_render()\n";
  }
  if ($t_count > 0) {
    $footer_doc_block_middle .= " *   $t_count call(s) to t(); can also use t as a filter in Twig\n";
  }
  if ($variables_ref_count > 0) {
    $footer_doc_block_middle .= " *   $variables_ref_count reference(s) to \$$variables_name; NB: just use the array key in the template (without trying to access $variables_name)\n";
  }
  $constant_string_count -= $variables_ref_count;
  if ($constant_string_count > 0) {
    $footer_doc_block_middle .= " *   approximately $constant_string_count strings of markup\n";
  }
  if (count($theme_calls) > 0) {
    $footer_doc_block_middle .= " *   " . count($theme_calls) . " call(s) to theme() found.  The function call should be removed and changed to simple arrays:\n";
    foreach ($theme_calls as $arguments_array) {
      $footer_doc_block_middle .= "array(\n  '#theme' => " . $arguments_array['hook'] . ",\n";
      if (isset($arguments_array['variables'])) {
        foreach ($arguments_array['variables'] as $key => $value) {
          $key = substr($key, 0, 1) . '#' . substr($key, 1);
          // @todo Hmm; var_export isn't quite right here either.
          $value = (gettype($value) === 'array') ? var_export($value, TRUE) : print_r($value, TRUE);
          $footer_doc_block_middle .= "  $key => $value,\n";
        }
      }
      $footer_doc_block_middle .= ")\n";
    }
  }
  $footer_docblock_end = " */\n";
  $footer_doc_block = $footer_doc_block_middle ? $footer_doc_block_start . $footer_doc_block_middle . $footer_docblock_end : '';

  return array(
    'function_text' => $header_doc_block . $output . $footer_doc_block,
    'function_name' => $new_function_name,
    'template_name' => $new_template_name,
    'template_token' => $template_token,
  );
}
